(*
  simple - a formal system demonstrating cyclic values
  Copyright (C) 1996  Scott Draves <spot@cs.cmu.edu>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundationeither version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this programif not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)
structure Syntax (* : SYNTAX *) =
struct

datatype Prim = p_plus | p_minus | p_times | p_div  | p_mod
              | p_equals | p_less | p_greater | p_less_equal | p_greater_equal
              | p_bit_and | p_bit_or | p_bit_xor | p_bit_not
              | p_bit_shift_left | p_bit_shift_right | p_aliases
              | p_load_word | p_store_word | p_effect
              | p_is_pair | p_lift | p_debug | p_pair | p_left | p_right
              | p_get_store | p_set_store | p_effect_store | p_not
              | p_is_dynamic

type myint = Word32.word

type Name = string

datatype Atom = atom_int of myint
              | atom_bool of bool
              | atom_unit

    datatype Exp = exp_atom of Atom
             | exp_var of Name
             | exp_app of {label:Name, arg:Exp, memoize:bool}
             | exp_loop of {index:M,  init:Exp, done:Exp, next:Exp, exit:Exp}
             | exp_let of {inline:bool, pattern:M, exp:Exp, body:Exp}
             | exp_if  of Exp * Exp * Exp
             | exp_and  of Exp * Exp
             | exp_or  of Exp * Exp
             | exp_prim of Prim * Exp list
             | exp_quote of Exp

             (* for output as C code only *)
             | exp_comma of Exp * Exp
             | exp_assign of Name * Exp
             | exp_while of Exp * Exp

             (*
             | exp_if2  of Exp * Exp * Exp
             | exp_abs of Name * Exp
             *)


and M      = m_static of Atom
           | m_dynamic of Exp
           | m_cyclic of myint * Exp * myint
           | m_pair of M * M
           | m_effect of Exp * M  (* deprecated *)

    type Version = {formal:M, store:M, label:Name, residual:(M*M)}
    type Versions = Version list ref

    datatype Def = def_fn of Name * M * Exp * Versions
                 | def_val of Name * Exp
                 | def_pat of Name * M

    type Program = Def list


fun minus (x:myint) = 0w0 - x

val minint = Word32.<<(0w1,0w31)

fun mydiv (x, y) = if (0w0 = y) then (print "div by zero"; raise Div)
                   else if (x < minint)
                   then x div y
                   else (minus ((minus x) div y))
                   

fun mymod (x, y) = if (0w0 = y) then (print "mod by zero"; raise Div)
                   else if (x < minint)
                   then x mod y
                   else (minus ((minus x) mod y))

fun unparse_val (atom_int i) =  if (i < minint) then (Word32.fmt StringCvt.DEC i)
                                else ("-" ^ (Word32.fmt StringCvt.DEC (minus i)))
  | unparse_val (atom_bool b) =  Bool.toString(b)
  | unparse_val (atom_unit) =  "unit"

fun unparse_prim p_plus =  "plus"
  | unparse_prim p_times =  "times"
  | unparse_prim p_equals =  "equals"
  | unparse_prim p_div =  "idiv"
  | unparse_prim p_mod =  "imod"
  | unparse_prim p_bit_and =  "bit_and"
  | unparse_prim p_bit_or =  "bit_or"
  | unparse_prim p_bit_xor =  "bit_xor"
  | unparse_prim p_bit_not =  "bit_not"
  | unparse_prim p_bit_shift_left =  "bit_shift_left"
  | unparse_prim p_bit_shift_right =  "bit_shift_right"
  | unparse_prim p_effect =  "effect"
  | unparse_prim p_load_word =  "load_word"
  | unparse_prim p_store_word =  "store_word"
  | unparse_prim p_minus =  "minus"
  | unparse_prim p_less =  "less"
  | unparse_prim p_greater =  "greater"
  | unparse_prim p_less_equal =  "less_equal"
  | unparse_prim p_greater_equal =  "greater_equal"
  | unparse_prim p_aliases =  "aliases"
  | unparse_prim p_is_pair =  "is_pair"
  | unparse_prim p_lift =  "lift"
  | unparse_prim p_debug =  "debug"
  | unparse_prim p_pair =  "pair"
  | unparse_prim p_left =  "left"
  | unparse_prim p_right =  "right"
  | unparse_prim p_get_store =  "get_store"
  | unparse_prim p_set_store =  "set_store"
  | unparse_prim p_effect_store =  "effect"
  | unparse_prim p_not =  "not"

fun unparse_list (hd :: tail) = " " ^ (unparse hd) ^ (unparse_list tail)
  | unparse_list _ = ""

and unparse (exp_atom v) =  unparse_val v
  | unparse (exp_var v) = v
  | unparse (exp_let{inline, pattern, exp, body}) = ( (if inline then "slet " else "dlet ")
                   ^ (unparse_m pattern) ^ " = " ^ (unparse exp) ^ "\nin " ^ (unparse body))
  | unparse (exp_app{label,arg,memoize}) = 
     ( "(" ^(Bool.toString memoize)^" "^ label ^ " " ^ (unparse arg) ^ ")")
  | unparse (exp_if(e0,e1,e2)) = ("(if " ^ (unparse e0) ^ " " ^ (unparse e1) ^ " " ^ (unparse e2) ^ ")")
  | unparse (exp_and(e0,e1)) = ("(and " ^ (unparse e0) ^  " " ^ (unparse e1) ^ ")")
  | unparse (exp_or(e0,e1)) = ("(and " ^ (unparse e0) ^  " " ^ (unparse e1) ^ ")")
  | unparse (exp_comma(e0,e1)) = ("(" ^ (unparse e0) ^  ", " ^ (unparse e1) ^ ")")
  | unparse (exp_prim(p,es)) = ( "(" ^ (unparse_prim p) ^ (unparse_list es) ^ ")")
  | unparse (exp_assign(v,e)) = ( "(" ^ v ^ " = " ^ (unparse e) ^ ")")
  | unparse (exp_while(e0,e1)) = "while (" ^ (unparse e0) ^ ") " ^ (unparse e1)
  | unparse (exp_quote(e0)) = "(quote " ^ (unparse e0) ^ ")"
  | unparse (exp_loop{index, init, done, next, exit}) = "loop "^(unparse_m index)^" "^
                       (unparse init)^" "^(unparse done)^" "^(unparse next)^" "
                       ^(unparse exit)


and unparse_m (m_static v) = ( "(s " ^ (unparse_val v) ^")")
  | unparse_m (m_dynamic d) = ( "(d " ^ (unparse d)^")")
  | unparse_m (m_cyclic(b,q,r)) = ("(c " ^ (unparse_val(atom_int b)) ^ " " ^ (unparse q) ^ " " ^ (unparse_val(atom_int r)) ^")")
  | unparse_m (m_pair(l,r)) = ( "(p " ^ (unparse_m l) ^ " "^ (unparse_m r) ^ ")")
  | unparse_m (m_effect(e,v)) = ( "(e " ^ (unparse e) ^ " " ^ (unparse_m v) ^ ")")

  fun ver {store,formal,label:Name, residual:(M*M)} = 
       (unparse_m store) ^ " " ^ label ^ " "^ (unparse_m formal) ^ " = "
     ^ (unparse_m (#1 residual)) ^ " " ^ (unparse_m (#2 residual)) ^ "\n"

  fun unparse_def (def_fn(name,pat,body,vs)) =
           "fun " ^ name ^ " " ^ (unparse_m pat) ^ " = " ^ (unparse body) ^ "\n" 
                ^ (concat (map ver (!vs)))
    | unparse_def (def_pat(name,m)) = "pat " ^ name ^ " = " ^ (unparse_m m) ^ "\n"
    | unparse_def (def_val(name,e)) = "val " ^ name ^ " = " ^ (unparse e) ^ "\n"

fun unparse_prog p = concat (map unparse_def p)


fun local_vars e = case e of
    (exp_let{inline, pattern, exp, body}) => local_vars(exp) @ local_vars(body) (* wrong XXX *)
  | (exp_atom _) => []
  | (exp_var _) => []
  | (exp_app {arg, ...}) => local_vars arg
  | (exp_if(e0,e1,e2)) => local_vars(e0) @ local_vars(e1) @ local_vars(e2)
  | (exp_prim(_,es)) => foldl op@ [] (map local_vars es)
  | (exp_comma(e0,e1)) => (local_vars e0) @ (local_vars e1)


fun reify (m_static v) = (exp_atom v)
  | reify (m_dynamic v) = v
  | reify (m_cyclic(b,q,r)) =
             (exp_prim(p_plus,[exp_prim(p_times,[q,exp_atom(atom_int(b))]),exp_atom(atom_int(r))]))
  | reify (m_pair(l,r)) = exp_prim(p_pair,[reify l, reify r])

fun reify_dynamic_part v = case v of
      (m_dynamic e) => SOME e
    | (m_cyclic(b,q,r)) => SOME q
    | (m_static(_)) => NONE
    | (m_pair(l,r)) => (case ((reify_dynamic_part l),(reify_dynamic_part r)) of
           (NONE, NONE) => NONE
         | (SOME x, NONE) => SOME x
         | (NONE, SOME x) => SOME x
         | (SOME x, SOME y) => SOME(exp_comma(x,y)))

fun flatten m =
   let val seen = ref []
       fun D e = if (List.exists (fn v => v=e) (!seen))
                            then NONE else (seen := (e::(!seen)); (SOME e))
       fun loop m = case m of
	   (m_dynamic e) => D e
	 | (m_cyclic(b,q,r)) => D q
	 | (m_static(_)) => NONE
	 | (m_pair(l,r)) => (case ((loop l),(loop r)) of
		(NONE, NONE) => NONE
	      | (SOME x, NONE) => SOME x
	      | (NONE, SOME x) => SOME x
	      | (SOME x, SOME y) => SOME(exp_comma(x,y)))
   in (loop m) end


local

  fun lv (exp_assign(v,e)) = v :: (lv e)
    | lv (exp_app{arg,...}) = lv arg
    | lv (exp_if(e0,e1,e2)) = (lv e0)@(lv e1)@(lv e2)
    | lv (exp_comma(e0,e1)) = (lv e0)@(lv e1)
    | lv (exp_prim(p,es)) = List.concat(map lv es)
    | lv (exp_while(p,e)) = (lv p)@(lv e)
    | lv (exp_and(e0,e1)) = (lv e0)@(lv e1)
    | lv (exp_or(e0,e1)) = (lv e0)@(lv e1)
    | lv (exp_let{pattern,exp,body,...}) = 
            (case pattern of (m_dynamic (exp_var n)) =>
		 n :: (lv exp)@(lv body))
    | lv _ = []

  fun uniq (x::xs) l = if (List.exists (fn v => v=x) l)
                       then uniq xs l
                       else uniq xs (x::l)
    | uniq _ l = l

  fun vars m = uniq (case (reify_dynamic_part m) of NONE => [] | SOME(e) => (lv e)) []

  fun um m = case (flatten m) of
               NONE => ""
             | SOME(e) => (ue e)

  and ue_wal (exp_and(p,q)) = (ue_wal p)@(ue_wal q)
    | ue_wal p = [p]

  and ue_while_ands_list e =
      concat (map (fn e => ("while(!("^(ue e)^"))")) (ue_wal e))

  and ue_val (atom_int i) = unparse_val(atom_int i)
    | ue_val (atom_bool true) = "c_true"
    | ue_val (atom_bool false) = "c_false"
    | ue_val (atom_unit) = "c_unit"

  and ue_list [e] = (ue e)
    | ue_list (e::tl) = (ue e) ^ ", " ^ (ue_list tl)
  and ue (exp_atom v) =  ue_val v
    | ue (exp_var v) = v
    | ue (exp_app{label,arg, ...}) = ( label ^ "(" ^ (ue arg) ^ ")")
    | ue (exp_if(e0,e1,e2)) = ("(" ^ (ue e0) ^ " ? " ^ (ue e1) ^ " : " ^ (ue e2) ^ ")")
    | ue (exp_comma(e0,e1)) = (ue e0) ^ "," ^ (ue e1)
    | ue (exp_prim(p,es)) = ((unparse_prim p) ^"(" ^ (ue_list es) ^ ")")
    | ue (exp_while(p,e)) = ("({"^(ue_while_ands_list p)^"{"^(ue e)^";}})")
    (* | ue (exp_while(exp_and(p,q),e)) = ("({while (!("^(ue p)^"))" ^
	                          "while (!("^(ue q)^")){"^(ue e)^";}})")
    | ue (exp_while(p,e)) = ("({while (!("^(ue p)^")){"^(ue e)^";}})") *)
    | ue (exp_assign(n,e)) = "("^ n ^ " = " ^ (ue e) ^ ")"
    | ue (exp_and(e0,e1)) = "("^ (ue e0) ^ "&&" ^ (ue e1) ^ ")"
    | ue (exp_or(e0,e1)) = "("^ (ue e0) ^ "||" ^ (ue e1) ^ ")"
    | ue (exp_let{pattern,exp,body,...}) =
      (case pattern of (m_dynamic x) 
	   => "(" ^ (ue x) ^ "=" ^ (ue exp) ^ "," ^ (ue body) ^ ")")
    | ue e = ((print ("not handled: "^ (unparse e)^"\n")); raise Match)
  fun commas [a] = a
    | commas (a::l) = a ^ ", " ^ (commas l)

  fun ver {formal, label, residual, store} =
       let val (rez,eff) = residual in (* lose effects XXX *)
         "word_t " ^ label ^ "(" ^ (um (m_pair(formal,store))) ^ ") {" ^
         "word_t " ^ (commas (vars rez)) ^ ";\n" ^
         "word_t _r = (" ^ (um rez) ^ ");\n" ^
         "return _r;\n}\n"
       end

  fun ud (def_fn(name,pat,body,ref vs)) = concat (map ver vs)
    | ud (def_pat(name, v)) = ""
    | ud (def_val(name, v)) = ""
in
  fun to_gcc p = concat(map ud p)
  val um = um
end

fun add_type t (m_pair (a, b)) = m_pair((add_type t a), (add_type t b))
  | add_type t (m_dynamic(exp_var nm)) = m_dynamic(exp_var(t^nm))
  | add_type t (m_cyclic(b,(exp_var nm),r)) = 
    m_cyclic(b,(exp_var(t^nm)),r)
  | add_type t x = x


end
